function [u, c, v, w, vtilde, wtilde] = FindMaskOld(u0, lambda, epsilon, mu, nMax, nOpt)
%% Data
[nr, nc] = size(u0);
nn = nr*nc;

%% Helper Functions
D1y = sparse(1:(nr-1),2:nr,1,nr,nr) - speye(nr, nr);
D1y(nr,nr) = 0;
D1x = sparse(1:(nc-1),2:nc,1,nc,nc) - speye(nc, nc);
D1x(nc,nc) = 0;

%% Operators
DxDy = [ kron(D1x,speye(nr,nr)) ;  kron(speye(nc,nc),D1y) ];   % Gradient
% Dxpy = [-kron(D1x,speye(nr,nr))', -kron(speye(nc,nc),D1y)']; % Divergence
Dxxyy = -DxDy'*DxDy;                                           % Laplacian

%% Primal Dual Solver with iterative linearisation
uk = u0(:);        % Initialisation (solution image)
ck = ones(nn, 1,'like',u0);  % Initialisation (solution mask)
f  = u0(:);        % Initialisation (image data)
%
ttime = 0;
for i = 1:nMax
    tic();
    %% Setup Matrices (linearisation)
    
    if isa(u0, 'gpuArray')
        % A = ck + (1-ck)(-L)
        A = gpuArray(sparse(1:nn, 1:nn, ck(:)) + ...
            (speye(nn,nn) - sparse(1:nn, 1:nn, ck(:)))*(-Dxxyy));
        
        % B = (uk - f + L*uk)
        B = gpuArray(sparse(1:nn, 1:nn, uk(:) - f(:) + Dxxyy*uk(:)));
        
        % ck (1 + L) uk
        g = gpuArray(ck(:).*uk(:) + ck(:).*(Dxxyy*uk));
    else
        % A = ck + (1-ck)(-L)
        A = sparse(1:nn, 1:nn, ck(:)) + ...
            (speye(nn,nn) - sparse(1:nn, 1:nn, ck(:)))*(-Dxxyy);
        
        % B = (uk - f + L*uk)
        B = sparse(1:nn, 1:nn, uk(:) - f(:) + Dxxyy*uk(:));
        
        % ck (1 + L) uk
        g = ck(:).*uk(:) + ck(:).*(Dxxyy*uk);
    end
    %% Primal Dual Solver
    y = zeros(size(uk),'like',uk);
    
    u    = uk(:);
    uhat = u;
    
    c    = ck(:);
    chat = c;
    
    L = max(abs(ck(:))) + max(1-abs(ck(:)))*4 + ...
        max(abs(uk(:) + f(:) - Dxxyy*uk(:)));
    % = 8; % = svds([A, B], 1);
    
    tau = 1/L^2;   % L^2
    sigma = 1;     % L^2
    
    theta = 1.0;
    acc = (epsilon + mu)/2;
    
    fprintf('Iteration %d:\n\n', i);
    tic();
    eOld = energy(u,c,f,uk,ck,lambda,mu,epsilon);
    %fprintf('energy (before) = %f\n\n', eOld);
    
    %fprintf('L = %s\n', L);
    %fprintf('tau/sigma (before) = %f - %f\n\n', tau, sigma);
    
    %fprintf('max(u) = %f\n',   max(uk(:)));
    %fprintf('min(u) = %f\n\n', min(uk(:)));
    
    %fprintf('max(c) = %f\n',   max(ck(:)));
    %fprintf('min(c) = %f\n\n', min(ck(:)));
    
    for j = 1:nOpt
        uold = u;
        cold = c;
        
        y = y + sigma * (A*uhat + B*chat - g);
        
        u = (u - tau*(A'*y-f-mu*uk))/(1+tau+mu*tau);
        
        c = shrink((c - tau*(B'*y) + tau*mu*ck)/(1 + tau*epsilon + tau*mu), ...
            tau*lambda/(1+tau*epsilon+tau*mu));
        
        if acc > 0
            theta = 1/sqrt(1 + 2*acc*tau);
            tau   = theta*tau;
            sigma = sigma/theta;
        end
        
        uhat = u + theta*(u - uold);
        chat = c + theta*(c - cold);
        
        if j > 20 && (norm(u-uold) < 1e-7 || norm(c-cold) < 1e-7)
            break % Stopping criteria primal dual
        end
      
    end % Inner Loop
    
    t = toc();
    disp(t);
    
    %disp([norm(u-uold), norm(c-cold)]);
    fprintf('Primal-Dual its: %d\n\n', j);
    %fprintf('tau/sigma (after) = %f - %f\n\n', tau, sigma);
    
    eNew = energy(u,c,f,uk,ck,lambda,mu,epsilon);
    %fprintf('energy (after) = %f\n', eNew);
    %fprintf('energy (diff)  = %f\n\n', eOld - eNew);
    %fprintf('Residual constraint = %f\n\n\n', norm(A*u+B*c-g));
    
    if norm(u-uk) < 1e-10 || norm(c-ck) < 1e-10 || abs(eOld - eNew) < 0.0001 || (eNew > eOld)
        break % Stopping criteria linearisations
    end
        
    %% Update
    uk = u(:);
    ck = c(:);
    
    time = toc();
    ttime = ttime + time;
    disp(['Time: ' num2str(ttime/i)]);
end % Outer Loop

A = sparse(1:nn, 1:nn, ck(:)) + ...
    (speye(nn,nn) - sparse(1:nn, 1:nn, ck(:)))*(-Dxxyy);

v = A\(ck(:).*f);
w = gvo(A, ck, f);

%err_v = norm(v - f)^2/numel(f);
%err_w = norm(w - f)^2/numel(f);

%fprintf('Before Binarisation\n');
%fprintf('MSE w/o GVO: %f\n', err_v);
%fprintf('MSE w/  GVO: %f\n', err_w);

%%%%%

ck(ck > 0.0001) = 1;
ck(ck < 0.9999) = 0;

A = sparse(1:nn, 1:nn, ck(:)) + ...
    (speye(nn,nn) - sparse(1:nn, 1:nn, ck(:)))*(-Dxxyy);

vtilde = A\(ck(:).*f);
wtilde = gvo(A, ck, f);

%err_v = norm(vtilde - f)^2/numel(f);
%err_w = norm(wtilde - f)^2/numel(f);

%fprintf('After Binarisation\n');
%fprintf('MSE w/o GVO: %f\n', err_v);
%fprintf('MSE w/  GVO: %f\n', err_w);

u = reshape(u, [nr, nc]);
c = reshape(c, [nr, nc]);

v = reshape(v, [nr, nc]);
w = reshape(w, [nr, nc]);

vtilde = reshape(vtilde, [nr, nc]);
wtilde = reshape(wtilde, [nr, nc]);

end

function y = shrink(x, l)
% Soft shrinkage with hard constraints x in [ul, uu].
y = zeros(size(x), 'like', x);
y(x >=  l) = x(x >=  l) - l;
y(x <= -l) = x(x <= -l) + l;
end

function e = energy(u,c,f,uk,ck,lambda,mu,epsi)
e = 0.5 * norm(u-f,2)^2 + ...
    lambda * norm(c,1) + epsi/2 * norm(c,2)^2 + ...
    mu/2 * norm(u-uk,2)^2 + mu/2 * norm(c-ck,2)^2;
end

function v = gvo(A, c, f)

dA = decomposition(A);

[g, flag] = lsqr(@afunA, f(:), 1.0e-10, 360);
if flag > 1
    error('Something went wrong!');
end

v = dA\(c(:).*g(:));

    function y = afunA(x,transp_flag)
        if strcmp(transp_flag,'transp')      % y = A'*x
            y = c(:).*((dA')\x);
        elseif strcmp(transp_flag,'notransp') % y = A*x
            y = dA\(c(:).*x);
        end
    end
end